''' only access the metadata location functions '''

import os
import hashlib

from datetime import datetime
from sqlitedict import SqliteDict
from .edgex_exceptions import *

# buffer hash computation
class EdgexHash:
    """ Calculate the hash for the buffer, file
        and pick up which hash algorithm to use
    """
    def __init__(self):
        pass

    @classmethod
    def signature(cls, file_name):
        """ Given a file calculate the signature """
        hash_sha256 = hashlib.sha256()
        file_des = open(file_name, 'rb')
        chunk = file_des.read()
        hash_sha256.update(chunk)
        return hash_sha256.hexdigest()

    @classmethod
    def md5(cls, databuf):
        """ Calculate the MD5 hash """
        hash_md5 = hashlib.md5()
        hash_md5.update(databuf)
        return hash_md5.hexdigest()

    @classmethod
    def sha256(cls, databuf):
        """ Calculate the SHA256 hash """
        hash_sha256 = hashlib.sha256()
        hash_sha256.update(databuf)
        return hash_sha256.hexdigest()


class EdgexMetaSQLite:
    """ Only store the meta data for the data """
    def __init__(self, cfg, vdb_file=None):
        """ COnstructor for the meta data store """
        self.path = cfg.get_meta_location()
        if vdb_file is None:
            vdb_file = "edgex_meta.db"
        self.mname = self.path + "/" + vdb_file

    def show(self):
        """ print out the metadata as it is -- debugging """
        metadict = SqliteDict(self.mname, autocommit=True)
        for key, value in metadict.items():
            print(key, value)
        metadict.close()
    def init_store(self):
        """ Initialize """
        metadict = SqliteDict(self.mname, autocommit=True)
        key = "edgex_meta"
        tm_now = datetime.utcnow()
        val = tm_now.strftime('%Y%m%dT%H%M%SZ')
        metadict[key] = val
        metadict.close()
    def put(self, object_path, databuf):
        """ create and put a meta entry for this object """
        hash_sha256 = EdgexHash()
        key = hash_sha256.sha256(databuf)
        val = object_path
        metadict = SqliteDict(self.mname, autocommit=True)
        metadict[key] = val
        metadict.close()

    def get(self, key):
        """ get the meta entry with this key """
        metadict = SqliteDict(self.mname, autocommit=True)
        retval = metadict[key]
        metadict.close()
        return retval

    def delete(self, key):
        """ delete this meta entry """
        metadict = SqliteDict(self.mname, autocommit=True)
        del metadict[key]
        metadict.commit()
        metadict.close()

    def clean_store(self):
        """ wipe out the meta db """
        if os.path.isfile(self.mname) is True:
            os.unlink(self.mname)



""" The main edgex_access file that has all the classes used in this
    module.
"""

from datetime import datetime
import os
import json
import random

import asyncio
import aiobotocore
import time

from logzero import logger

from .edgex_exceptions import *
from .edgex_base import *
from .edgex_fs import *
from .edgex_s3 import *
from .edgex_mem import *
from .edgex_meta import *
from .edgex_goog import *

# import requests-aws4auth

# buffer hash computation

class EdgexHash:
    """ Calculate the hash for the buffer, file
        and pick up which hash algorithm to use
    """
    def __init__(self):
        pass

    @classmethod
    def signature(cls, file_name):
        """ Given a file calculate the signature """
        hash_sha256 = hashlib.sha256()
        file_des = open(file_name, 'rb')
        chunk = file_des.read()
        hash_sha256.update(chunk)
        return hash_sha256.hexdigest()

    @classmethod
    def md5(cls, databuf):
        """ Calculate the MD5 hash """
        hash_md5 = hashlib.md5()
        hash_md5.update(databuf)
        return hash_md5.hexdigest()

    @classmethod
    def sha256(cls, databuf):
        """ Calculate the SHA256 hash """
        hash_sha256 = hashlib.sha256()
        hash_sha256.update(databuf)
        return hash_sha256.hexdigest()


class EdgexStore:
    """ Specific description of each store """

    def __init__(self, cfg):
        if cfg['STORE_TYPE'] == "S3":
            self.store = EdgexS3Store(cfg)
        elif cfg['STORE_TYPE'] == "FS":
            self.store = EdgexFSStore(cfg)
        elif cfg['STORE_TYPE'] == "MEM":
            self.store = EdgexMemStore(cfg)
        else:
            raise InvalidStore(cfg['STORE_TYPE'])

    def islocal(self):
        """ return if this is a local store """
        return self.store.islocal()

    def get_access(self):
        """ return the access key """
        return self.store.get_access()

    def get_secret(self):
        """ return the secret key """
        return self.store.get_secret()

    def get_region(self):
        """ return the region """
        return self.store.get_region()

    def get_endpoint(self):
        """ return the end point fot the store """
        return self.store.get_endpoint()

    def get_name(self):
        """ return the name """
        return self.store.get_name()

    def get_type(self):
        """ return the type """
        return self.store.get_type()

    def default_bucket(self):
        """ the default bucket """
        return self.store.default_bucket()
    def change_value(self, key, value):
        """ change the internal config value """
        return self.store.change_value(key, value)

class EdgexConfig:
    """ Read the main config for the EdgexAccess """
    file_name = ""

    def __init__(self, cfg_filename):
        self.prog = "SP3"
        with open(cfg_filename, "r") as open_cfg:
            self.cfg_data = json.load(open_cfg)

        self.file_name = cfg_filename
        self.store_dict = {}
        stores = self.cfg_data['stores']
        for st_name in stores:
            self.store_dict[st_name['NAME']] = EdgexStore(st_name)

        if self.cfg_data['PRIMARY']:
            self.primary = self.cfg_data['PRIMARY']
        if self.cfg_data['DEBUG']:
            self.debug_level = self.cfg_data['DEBUG']
        if self.cfg_data['SYNCIO']:
            self.syncio = self.cfg_data['SYNCIO']
        if self.cfg_data['DATATEST']:
            self.datatest = self.cfg_data['DATATEST']
        if self.cfg_data['META']:
            self.meta = self.cfg_data['META']

        # setup the pwd store and the in-memory store
        pname = os.getcwd().split("/")
        ep = "/".join(pname[0:-2])
        store = self.create_store("pwd", "FS", pname[-1], endpoint=ep, tag="pwd")
        self.add_store("pwd", store)
        ep ="/sqlite/"
        store = self.create_store("mem", "MEM", str(os.getpid()), \
                                  endpoint=ep, tag="mem")
        self.add_store("MEM", store)

    def io_type(self):
        """ whether this is sync or async """
        return self.syncio

    def change_cfg_value(self, key, values):
        """ edit the key for this config """
        if len(values) > 0:
            self.cfg_data[key.upper()] = values[0]
        else:
            raise InvalidArgument(key)

    def change_store_value(self, st_name, var_name, var_value):
        """ Change one value in the sore """
        stores = self.cfg_data['stores']
        for sname in stores:
            if sname['NAME'] == st_name:
                stcfg = self.store_dict[sname['NAME']]
                stcfg.change_value(var_name, var_value)
                sname[var_name.upper()] = var_value

    def create_store(self, name, store_type, bucket, \
               access="", secret="", endpoint=None, \
               region="", token="", tag=""):
        """ create a store instance """

        if not name or not store_type or not bucket:
            raise InvalidArgument(name)

        scfg = {}
        scfg['NAME'] = name
        scfg['STORE_TYPE'] = store_type
        scfg['BUCKET'] = bucket

        # optional
        if store_type == 'S3':
            scfg['ACCESS'] = access
            scfg['SECRET'] = secret

        # optional
        scfg['ENDPOINT'] = endpoint
        scfg['REGION'] = region
        scfg['TOKEN'] = token
        scfg['TAG'] = tag
        

        # jcfg = json.dumps(scfg)
        if store_type == "FS":
            store = EdgexFSStore(scfg)
        elif store_type == "S3":
            store = EdgexS3Store(scfg)
        elif store_type == "MEM":
            store = EdgexMemStore(scfg)
        else:
            raise InvalidStore(store_type)

        return store

    def get_store(self, store_name):
        """ get me a store with this name from the dictionary """
        try:
            store = self.store_dict[store_name]
            return store
        except Exception as exp:
            logger.exception(exp)
            raise InvalidStore(store_name)

    def add_store(self, name, store):
        """ add this store object with a name as a key """
        self.store_dict[name] = store
        st_obj = {}
        st_obj['NAME'] = store.get_name()
        st_obj['STORE_TYPE'] = store.get_type()
        st_obj['BUCKET'] = store.default_bucket()
        st_obj['ENDPOINT'] = store.get_endpoint()
        st_obj["TOKEN"] = ""
        st_obj["TAG"] = ""
        if store.get_type() == "S3":
            st_obj["ACCESS"] = ""
            st_obj["SECRET"] = ""
            st_obj["REGION"] = ""
            st_obj["ENDPOINT"] = ""
        self.cfg_data['stores'].append(st_obj)

    def del_store(self, name):
        """ delete this store object with a name as a key """
        del self.store_dict[name]
        for k in self.cfg_data['stores']:
            if k['NAME'] == name:
                del k

    def get_pwd_store(self):
        """ return a store corresponding to the pwd in local store """
        return self.store_dict['pwd']

    def get_mem_store(self):
        """ return a store corresponding to the pwd in local store """
        return self.store_dict['mem']

    def get_primary_store(self):
        """ the name of the primary store """
        if self.cfg_data['PRIMARY'] is None:
            raise RuntimeError("No Primary Store defined")
        return self.store_dict[self.cfg_data['PRIMARY']]

    def get_meta_location(self):
        """ the location of metadata is chose to be stored """
        if self.cfg_data['META'] is None:
            raise RuntimeError("No Meta Store defined")
        return self.cfg_data['META']

    def get_datatest(self):
        """ this is the test/ directory """
        cwd = os.getcwd()
        dpath = cwd + "/" + self.datatest
        if not os.path.exists(dpath):
            os.makedirs(dpath)
        return dpath

    def show_store(self, store):
        """ show one store """
        logger.info(str("\t" + store.get_name() + "\t" + store.get_type() \
                            + "\t" + store.default_bucket()))

    def show_stores(self):
        """ show me all the stores """
        for k in self.store_dict:
            store = self.store_dict[k]
            self.show_store(store)

    def show_all(self):
        """ show all the stores """
        logger.info(str("primary:" + "\t" + self.primary))
        logger.info(str("debug_level: " + "\t" + str(self.debug_level)))
        logger.info(str("syncio :" + "\t" + self.syncio))
        logger.info("stores:")
        self.show_stores()

    def save(self):
        """ convert to a json format for saving """
        cfg_obj = {}
        cfg_obj["PROGRAM"] = self.prog
        for k in self.cfg_data:
            cfg_obj[k] = self.cfg_data[k]
            if k == 'stores':
                c = 0
                for st in cfg_obj[k]:
                    if st['NAME'] == "mem":
                        del cfg_obj[k][c]
                    if st['NAME'] == "pwd":
                        del cfg_obj[k][c]
                    c += 1
        with open(self.file_name, 'w') as outfile:
            json.dump(cfg_obj, outfile, indent=4, sort_keys=True)

class EdgexObjectName:
    """ Only define the name string of the obeject and
        parse out the store, bucket, objname, if it is a
        folder etc from it
        examples of names:

        aws3://xenocloud/foo/bar
            store is defined in the cfg as aws3 with all the credentials
            bucketname is xenocloud and the path tp the object bar is /foo/bar

        ix://mydir/foo/bar
            store is defined in the config as ix, as a Filesystem path
            bucket name ( folder name ) is mydir and the path to the file bar
            is /foo/bar

        The following two stes are automatically created and are not in the
        config

        mem://2324/stuff
            store is an im-memory store for pid 2324. Object is stuff.

        pwd://mydir/foo/bar
            store is the present working directory. Folder is mydir
            and the path to the object bar is /foo/bar

    """
    def __init__(self, name):
        if len(name) == 0:
            raise InvalidStore("Store not defined :" + name)
        self.oname = name
        self.objname = ""
        self.bucketname = ""
        self.storename = ""
        self.rpath = ""
        self.storename, self.rpath = self.parse_storename()
        self.bucketname, self.objname = self.parse_bucketobject()
        logger.debug(str("Store: " + self.storename + "\tBucket: " \
                         + self.bucketname + "\tObject: " \
                         + self.objname))

    def parse_storename(self):
        """ returns the store name and the remaining path
            the oname must be of the form
            store://bucket/obj/name/path
        """
        sname = self.oname.split(":")
        storename = ""
        rpath = ""
        if len(sname) == 2:
            storename = sname[0]
            rpath = sname[1]
        elif len(sname) == 1:
            # the format store://bucket/obj was not used
            rpath = sname[0]
        else:
            raise InvalidStore("Store not defined: " + sname[0])
        return storename, rpath

    def parse_bucketobject(self):
        """ returns the bucketname, objectname """
        bname = self.rpath.split("/")
        bucketname = ""
        objectname = ""
        if len(bname) > 1:
            bucketname = bname[2]
            objectname = "/".join(bname[3:])
            return bucketname, objectname
        raise InvalidBucketName(self.rpath)

    def isfolder(self):
        """ determine if this is a folder by the name only """
        return self.objname.endswith("/")

    def get_storename(self):
        """ return only the store name """
        return self.storename

    def get_objectname(self):
        """ return only the object name """
        return self.objname

    def get_bucketname(self):
        """ return only the bucket name """
        return self.bucketname


class EdgexObject:
    """ defines the main object as defined by the object name """

    def __init__(self, cfg, name, store=None):
        self.is_object = False
        self.cfg = cfg
        # contains the databuffer on one task only. .. not the entire content-length.
        self.databuf = None
        # used only to pass around in callbacks etc
        self.arg = None
        self.ctx = None

        # the entire thing as the name says
        self.pjname = EdgexObjectName(name)

        # the object name only
        self.obj_name = self.pjname.get_objectname()
        self.bucket_name = self.pjname.get_bucketname()

        if store is not None:
            self.store = store
        else:
            self.store = cfg.get_store(self.pjname.get_storename())
            if self.store is None:
                raise InvalidStore(store)

        if self.store is None:
            self.store = self.cfg.get_primary_store()
        if not self.bucket_name:
            logger.debug("No Bucket name")
            self.bucket_name = self.store.default_bucket()

        # time for the creation of this in-memory object
        tm_now = datetime.utcnow()
        self.amzdate = tm_now.strftime('%Y%m%dT%H%M%SZ')
        self.datestamp = tm_now.strftime('%Y%m%d') # Date w/o time, used in credential scope

        #logger.debug(str("OBJECT : " + self.pathname() + "\t" + str(self.datestamp)))
        logger.debug(str("OBJECT : " + self.pathname()))

    def treat_object(self):
        """ no recursion, just treat this as os an object plain as simple 
            even if it has the "/" suffix """
        self.is_object = True

    def isfolder(self):
        """ return of the entire object is a folder
            or if the bucket is only there
            or if only the store is listed
        """
        if self.is_object:
            return False

        if self.store.get_type() == "FS":
            return os.path.isdir(self.pathname())
        if self.obj_name:
            return self.pjname.isfolder()
        if self.bucket_name:
            return self.bucket_name.endswith("/")
        return False

    def islocal(self):
        """ return if this is a local object """
        if self.store == "FS":
            return os.path.isdir(self.pathname())
        return False

    def random_buffer(self):
        """ Initialize this with a random buffer from 1K - 8K """
        ksize = [1024, 2048, 4096, 8192]
        size_sz = random.choice(ksize)
        self.databuf = os.urandom(size_sz)

    def add_buffer(self, buffer):
        """ use this as a data buffer ... usually strings .."""
        self.databuf = buffer


    def get_store(self):
        """ return the store for this object """
        return self.store

    def store_type(self):
        """ return the type if available """
        return self.store.get_type()

    def bucketname(self):
        """ return the bucket name """
        return self.bucket_name

    def objname(self):
        """ return the object name """
        return self.obj_name

    def basename(self):
        """ return only the storename://bucketname of this object """
        if self.store.get_name() != "local":
            fpath = self.store.get_name() + "://" + self.bucket_name + "/"
        else:
            fpath = self.store.get_name() + ":/" + self.bucket_name + "/"
        return fpath

    def stat(self, create=False):
        """ For local cases see if it is there, else create a dir path """
        if self.store_type() == "FS":
            file_found = os.path.exists(self.pathname())
            if (file_found is False) and (create is True) and self.obj_name.endswith("/"):
                logger.info(str("mkdir " + self.pathname()))
                os.makedirs(self.pathname())
            else:
                return file_found
        else:
            logger.error(str("Error: No stat on store_type: " + self.store_type()))
            raise InvalidStore(str(self.store_type()))
        return ""

    def pathname(self):
        """ return the full path name of the object """
        if self.store_type() == "FS":
            fpath = self.store.get_endpoint() + "/" + self.bucket_name + "/" + self.obj_name
        elif self.store_type() == "S3":
            fpath = self.store.get_endpoint() + "/" + self.bucket_name + "/" + self.obj_name
        elif self.store_type() == "MEM":
            fpath = self.bucket_name + "/" + self.obj_name
        else:
            logger.error(str("Error: store_type: " + self.store_type()))
            raise InvalidStore(str(self.store_type()))
        return fpath

    # TODO: remove this later
    def auth(self):
        # auth = AWS4Auth(self.store.access, self.store.secret, self.store.region, 's3')
        auth = ""
        return auth

    # return only the name
    def addchild(self, child):
        """ create a child on this object """
        objname = self.store.get_name() + "://" + \
                    self.bucketname() + "/" + self.objname() + \
                    child
        childobj = EdgexObject(self.cfg, objname)
        return childobj


class EdgexAccess:
    """ the main edgex_access class with the main
        methods to list, get, put, delete etc
    """
    def __init__(self, obj):
        if obj is None:
            raise InvalidArgument(str(None))
        self.obj = obj

        store_t = obj.store_type()
        if store_t == "FS":
            self.obj_access = EdgexFSAccess(obj)
        elif store_t == "S3":
            self.obj_access = EdgexS3Access(obj)
        elif store_t == "MEM":
            self.obj_access = EdgexMemAccess(obj)
        else:
            raise InvalidStore(self.obj.store_type())

    async def list(self, session=None):
        """ List the elements in this folder """
        logger.info(str("list " + self.obj.pathname()))
        return await self.obj_access.list(session)
    async def put(self, session=None):
        """ put this object using the desired access """
        logger.info(str("put " + self.obj.pathname()))
        return await self.obj_access.put(session)
    async def get(self, session=None):
        """ get this object using the desired access """
        logger.info(str("get " + self.obj.pathname()))
        return await self.obj_access.get(session)
    async def delete(self, session=None):
        """ delete this object using the desired access """
        logger.info(str("delete " + self.obj.pathname()))
        return await self.obj_access.delete(session)
    async def exists(self, session=None):
        """ exists this object using the desired access """
        logger.info(str("exists " + self.obj.pathname()))
        return await self.obj_access.exists(session)
    async def info(self, session=None):
        """ info this object using the desired access """
        logger.info(str("info " + self.obj.pathname()))
        return await self.obj_access.info(session)

class EdgexMeta:
    """ This object is specific to meta data only """
    def __init__(self, cfg, vdb):
        """ Initialize """
        self.obj = EdgexMetaSQLite(cfg, vdb_file=vdb)
    def init(self):
        """ Initialize """
        self.obj.init_store()
    def put(self, key, value):
        """ put a key value to this meta store """
        self.obj.put(key, value)
    def get(self, key):
        """ retrieve the value for this key """
        return self.obj.get(key)
    def delete(self, key):
        """ delete the  element with this key """
        self.obj.delete(key)
    def show(self):
        """ display what we have """
        self.obj.show()
    def wipe(self):
        """ wipe out the entire store"""
        self.obj.clean_store()


class EdgexDataAccess:
    """ main data access methods """
    cfg = None
    tasks = []

    def __init__(self, cfg, loop):
        self.cfg = cfg
        self.session = aiobotocore.get_session(loop=loop)
        self.loop = loop

    async def copy(self, source_obj, dest_obj, recursive=False):
        """ Copy one object from one store to another """
        if not source_obj or not dest_obj:
            raise InvalidArgument
        if recursive and source_obj.isfolder() and dest_obj.isfolder():
            contents = await self.ls(source_obj)
            logger.info(contents)
            for item in contents:
                child_source = source_obj.addchild(item)
                child_dest = dest_obj.addchild(item)
                self.tasks.append(await self.copy(child_source, child_dest, \
                                               child_source.isfolder()))
        else:
            try:
                source_obj.arg = dest_obj
                edgex_op = EdgexAccess(source_obj)
                databuf = await edgex_op.get(self.session)
                await self.gp_callback(self.session, 'get', source_obj, databuf)
                return True
            except Exception as exp:
                logger.exception(exp)
        return False

    async def delete(self, source_obj, recursive=False, parent_obj=None):
        """ Delete the object from the store """
        if not source_obj:
            raise InvalidArgument
        if source_obj.isfolder():
            if recursive:
                contents = await self.ls(source_obj)
                if len(contents) == 0:
                    source_obj.treat_object()
                    self.tasks.append(await self.delete(source_obj, \
                    #asyncio.ensure_future(await self.delete(source_obj, \
                                                  recursive=False, \
                                                  parent_obj=source_obj))
                for item in contents:
                    child_source = source_obj.addchild(item)
                    #asyncio.create_task(self.delete(child_source, \
                    self.tasks.append(await self.delete(child_source, \
                    #asyncio.ensure_future(await self.delete(child_source, \
                                                      child_source.isfolder(), \
                                                      parent_obj=source_obj))
        else:
            try:
                edgex_op = EdgexAccess(source_obj)
                deleted = await edgex_op.delete(self.session)
                await self.cmd_callback('delete', source_obj, deleted)
            except Exception as exp:
                logger.exception(exp)
        if parent_obj:
            contents = await self.ls(parent_obj)
            if len(contents) == 0:
                parent_obj.treat_object()
                edgex_op = EdgexAccess(parent_obj)
                deleted = await edgex_op.delete(self.session)
                await self.cmd_callback('delete', parent_obj, deleted)

    async def ls(self, source_obj):
        """ List the object or a folder """
        try:
            edgex_op = EdgexAccess(source_obj)
            list_out = await edgex_op.list(self.session)
            contents = await self.cmd_callback('list', source_obj, list_out)
            return contents
        except Exception as exp:
            logger.exception(exp)
        return []

    async def get(self, source_obj, dest_obj):
        """ Get the object from the store """
        if not source_obj or not dest_obj:
            raise InvalidArgument
        try:
            source_obj.arg = dest_obj
            edgex_op = EdgexAccess(source_obj)
            databuf = await edgex_op.get(self.session)
            await self.gp_callback(self.session, 'get', source_obj, databuf)
            return True
        except Exception as exp:
            logger.exception(exp)
        return False
    async def put(self, source_obj, dest_obj):
        """ put the object into the store """
        if not source_obj or not dest_obj:
            raise InvalidArgument
        try:
            source_obj.arg = dest_obj
            edgex_op = EdgexAccess(source_obj)
            databuf = await edgex_op.put(self.session)
            await self.gp_callback(self.session, 'put', source_obj, databuf)

        except Exception as exp:
            logger.exception(exp)
        return False

    async def exists(self, source_obj):
        """ Check if this object exists in the store """
        if not source_obj:
            raise InvalidArgument
        try:
            edgex_op = EdgexAccess(source_obj)
            is_there = await edgex_op.exists(self.session)
            await self.cmd_callback('exists', source_obj, is_there)
            return is_there
        except Exception as exp:
            logger.exception(exp)
        return False

    async def info(self, source_obj):
        """ Get the metadata for the object as seen by the store """
        if not source_obj:
            raise InvalidArgument
        try:
            edgex_op = EdgexAccess(source_obj)
            obj_info = await edgex_op.info(self.session)
            await self.cmd_callback('info', source_obj, obj_info)
            return obj_info
        except Exception as exp:
            logger.exception(exp)
        return None

    async def gend(self, dest_obj):
        """ artificial random data generation """
        if not dest_obj:
            raise InvalidArgument
        try:
            # the genfile part of ofile should be the lastpart of the dest_obj name
            ofile = "MEM://" + str(os.getpid()) + "/genfile"
            mem_source_obj = EdgexObject(self.cfg, ofile)
            mem_source_obj.random_buffer()
            mem_source_obj.arg = dest_obj
            edgex_op = EdgexAccess(mem_source_obj)
            databuf = await edgex_op.get(self.session)
            await self.gp_callback(self.session, 'put', mem_source_obj, databuf)
            # now delete the in-memory object
            await edgex_op.delete(self.session)
        except Exception as exp:
            logger.exception(exp)
        return None

    async def cmd_callback(self, cmd, obj, result):
        """ single command callback """
        #logger.info(str(result))
        if cmd == 'list':
            return result
        if not obj:
            pass

    async def gp_callback(self, session, cmd, obj, result):
        """ get-put callback """
        try:
            if not cmd:
                pass
            dest_obj = obj.arg
            dest_obj.databuf = result
            edgex_op = EdgexAccess(dest_obj)
            put_obj = await edgex_op.put(session)
            await self.cmd_callback('put', dest_obj, put_obj)
        except Exception as exp:
            logger.exception(exp)

    def create_task(self, task_name, source_name, dest_name, recursive=False):
        """ Create a particular type of task or tasks if recursive """
        if task_name == 'gend':
            if recursive:
                gend_tcount = 60 # total
                gend_mcount = 10 # modcount
                mdpath = source_name + "/d0"
                for i in range(0, gend_tcount):
                    rfile = mdpath + "/dd" + str(i)
                    source_obj = EdgexObject(self.cfg, rfile)
                    self.tasks.append(self.gend(source_obj))
                    if (i % gend_mcount) == 0 and (i != 0):
                        mdpath = source_name + "/" + "d" + str(i)
            else:
                source_obj = EdgexObject(self.cfg, source_name)
                self.tasks.append(self.gend(source_obj))
        elif task_name == 'ls':
            source_obj = EdgexObject(self.cfg, source_name)
            self.tasks.append(self.ls(source_obj))
        elif task_name == 'exists':
            source_obj = EdgexObject(self.cfg, source_name)
            self.tasks.append(self.exists(source_obj))
        elif task_name == 'info':
            source_obj = EdgexObject(self.cfg, source_name)
            self.tasks.append(self.info(source_obj))
        elif task_name == 'put':
            source_obj = EdgexObject(self.cfg, source_name)
            dest_obj = EdgexObject(self.cfg, dest_name)
            self.tasks.append(self.put(source_obj, dest_obj))
        elif task_name == 'get':
            source_obj = EdgexObject(self.cfg, source_name)
            dest_obj = EdgexObject(self.cfg, dest_name)
            self.tasks.append(self.get(source_obj, dest_obj))
        elif task_name == 'delete':
            source_obj = EdgexObject(self.cfg, source_name)
            self.tasks.append(self.delete(source_obj, recursive))
        elif task_name == 'copy':
            source_obj = EdgexObject(self.cfg, source_name)
            dest_obj = EdgexObject(self.cfg, dest_name)
            self.tasks.append(self.copy(source_obj, dest_obj, recursive))
        else:
            raise InvalidArgument(task_name)


    def execute(self):
        """ kick off the task list into the event q """
        self.loop.run_until_complete(asyncio.gather(*self.tasks))



''' Base classes use by higher level '''

from .edgex_exceptions import *

class EdgexStoreBase:
    """ Base class for Store definition """
    def __init__(self, cfg):
        self.name = cfg['NAME']
        self.type = cfg['STORE_TYPE']
        self.bucket = cfg['BUCKET']
        self.token = cfg['TOKEN']
        self.tag = cfg['TAG']
        #if not self.tag:
            #raise EmptyTag(self.tag)

    def islocal(self):
        """ is this a local filesystem """
        return self.type == "FS"
    def get_tag(self):
        """ tag is the short name we use """
        return self.tag
    def set_tag(self, avalue):
        """ set tag value """
        self.tag = avalue
    def get_name(self):
        """ Get the name """
        return self.name
    def get_type(self):
        """ Get the type """
        return self.type
    def default_bucket(self):
        """ Get the named default bucket """
        return self.bucket

class EdgexMetaBase:
    """ Base class to access the metadata only """
    state = ""
    def __init__(self):
        self.store_file = ""
    def init_store(self, store_file):
        """ Initialize the meta store if needed """
        self.store_file = store_file
        self.state = "init"
    def clear_store(self):
        """ remove and delete the meta store """
        self.state = "clear"

class EdgexAccessBase:
    """ Base class to access the I/O type. FS, S3 ,In-Mem etc """
    def __init__(self, obj, atype):
        self.obj = obj
        self.type = atype
    def get_type(self):
        """ if this is S3, FS, MEM """
        return self.type
    def get_obj(self):
        """ return the obj code """
        return self.obj



import os
import sys

path = os.path.join(__path__[0], 'version.py')
__version__ = eval(open(path).read())
__all__ = ['EdgexAccess', 'EdgexConfig', 'EdgexStore', 'EdgexObject', 'EdgexHash',
           'EdgexStoreBase', 'EdgexAccessBase',
           'EdgexFSStore',
           'EdgexFSAccess',
           'EdgexGoogStore',
           'EdgexGoogAccess',
           'EdgexMemStore',
           'EdgexMemAccess',
           'EdgexMeta',
           'EdgexS3Store',
           'EdgexS3Access',
           # errors
           'EdgexException',
           'AccessDenied',
           'AccountProblem',
           'AmbiguousGrantByEmailAddress',
           'BadDigest',
           'BucketAlreadyExists',
           'BucketAlreadyOwnedByYou',
           'BucketNotEmpty',
           'CredentialsNotSupported',
           'CrossLocationLoggingProhibited',
           'EntityTooSmall',
           'EntityTooLarge',
           'ExpiredToken',
           'IllegalVersioningConfigurationException',
           'IncompleteBody',
           'IncorrectNumberOfFilesInPostRequest',
           'InlineDataTooLarge',
           'InternalError',
           'InvalidAccessKeyId',
           'InvalidAddressingHeader',
           'InvalidArgument',
           'InvalidBucketName',
           'InvalidBucketState',
           'InvalidDigest',
           'InvalidEncryptionAlgorithmError',
           'InvalidLocationConstraint',
           'InvalidObjectState',
           'InvalidPart',
           'InvalidPartOrder',
           'InvalidPayer',
           'InvalidPolicyDocument',
           'InvalidRange',
           'InvalidRequest',
           'InvalidSecurity',
           'InvalidSOAPRequest',
           'InvalidStorageClass',
           'InvalidTargetBucketForLogging',
           'InvalidToken',
           'InvalidURI',
           'InvalidCommand',
           'InvalidStore',
           'InvalidDataBuffer',
           'KeyTooLong',
           'MalformedACLError',
           'MalformedPOSTRequest',
           'MalformedXML',
           'MaxMessageLengthExceeded',
           'MaxPostPreDataLengthExceededError',
           'MetadataTooLarge',
           'MethodNotAllowed',
           'MissingAttachment',
           'MissingContentLength',
           'MissingRequestBodyError',
           'MissingSecurityElement',
           'MissingSecurityHeader',
           'NoLoggingStatusForKey',
           'NoSuchBucket',
           'NoSuchKey',
           'NoSuchLifecycleConfiguration',
           'NoSuchUpload',
           'NoSuchVersion',
           'NotSignedUp',
           'NotSuchBucketPolicy',
           'OperationAborted',
           'PermanentRedirect',
           'PreconditionFailed',
           'Redirect',
           'RestoreAlreadyInProgress',
           'RequestIsNotMultiPartContent',
           'RequestTimeout',
           'RequestTimeTooSkewed',
           'RequestTorrentOfBucketError',
           'SignatureDoesNotMatch',
           'ServiceUnavailable',
           'SlowDown',
           'TemporaryRedirect',
           'TokenRefreshRequired',
           'TooManyBuckets',
           'UnexpectedContent',
           'UnresolvableGrantByEmailAddress',
           'UserKeyMustBeSpecified',
           'EmptyTag']

from .edgex_exceptions import *
from .edgex_base import *
from .edgex_fs import *
from .edgex_s3 import *
from .edgex_mem import *
from .edgex_goog import *
from .edgex_access import *





''' S3 specific functions '''

import os
import async_timeout

from logzero import logger

from .edgex_exceptions import *
from .edgex_base import EdgexStoreBase, EdgexAccessBase

class EdgexS3Store(EdgexStoreBase):
    """ Security credentials to access the
        store
    """
    def __init__(self, cfg):
        super().__init__(cfg)
        self.access = cfg['ACCESS']
        self.secret = cfg['SECRET']
        self.region = cfg['REGION']
        self.endpoint = cfg['ENDPOINT']
        self.tag = cfg['TAG']

    def get_endpoint(self):
        """ return the declared endpoint """
        return self.endpoint
    def set_endpoint(self, avalue):
        """ set the endpoint value """
        self.endpoint = avalue
    def get_region(self):
        """ the default region is needed for AWS """
        return self.region
    def set_region(self, avalue):
        """ set the region value """
        self.region = avalue
    def get_access(self):
        """ The ACCESS value """
        return self.access
    def set_access(self, avalue):
        """ set the access value """
        self.access = avalue
    def get_secret(self):
        """ TheSECRET vaue """
        return self.secret
    def set_secret(self, avalue):
        """ set the secret value """
        self.secret = avalue
    def get_tag(self):
        """ TheTAG vaue """
        return self.tag
    def set_tag(self, avalue):
        """ set the tag value """
        self.tag = avalue
    def change_value(self, key, value):
        """ change the config value """
        if key == "ACCESS":
            self.set_access(value)
        if key == "SECRET":
            self.set_secret(value)
        if key == "REGION":
            self.set_region(value)
        if key == "ENDPOINT":
            self.set_endpoint(value)
        if key == "TAG":
            self.set_tag(value)

class EdgexS3Access(EdgexAccessBase):
    """ S3 protocol access """
    def __init__(self, obj):
        super().__init__(obj, "S3")

    async def list(self, session=None):
        """ List the elements """
        final_list = []
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                            aws_secret_access_key=self.obj.store.get_secret(), \
                            aws_access_key_id=self.obj.store.get_access(), \
                            endpoint_url=self.obj.store.get_endpoint()) as client:

            prefix = self.obj.objname()
            resp = await client.list_objects(Bucket=self.obj.bucketname(), \
                                             Prefix=prefix, Delimiter='/')
            retcode = resp['ResponseMetadata']['HTTPStatusCode']
            if retcode != 200:
                raise RuntimeError("HTTP Error {}".format(retcode))

            if 'CommonPrefixes' in resp:
                for r_x in resp['CommonPrefixes']:
                    if prefix.endswith('/') and prefix and (prefix != r_x['Prefix']):
                        final_list.append(r_x['Prefix'].replace(prefix, ''))
                    elif not prefix:
                        final_list.append(r_x['Prefix'])
                    else:
                        dlist = r_x['Prefix'].split('/')
                        if dlist:
                            if len(dlist[-1]) > 0:
                                final_list.append(dlist[-1])
                            elif len(dlist[-2]) > 0:
                                final_list.append(dlist[-2])
                        else:
                            final_list.append(r_x['Prefix'])
            elif 'Contents' in resp:
                for r_x in resp['Contents']:
                    if prefix.endswith('/') and prefix:
                        final_list.append(r_x['Key'].replace(prefix, ''))
                    else:
                        dlist = r_x['Key'].split('/')
                        if dlist:
                            final_list.append(dlist[-1])
                        else:
                            final_list.append(r_x['Prefix'])
            logger.debug("s3 list : " + str(len(final_list)) + " items")
            return final_list

    async def put(self, session=None):
        """ PUT the buffer in databuf using the S3protocol """
        isdbuf = (self.obj.databuf is not None)
        logger.info(str("put " + self.obj.pathname() + \
                        " databuf " + str(isdbuf)))
        if not isdbuf:
            try:
                os.makedirs(os.path.dirname(self.obj.pathname()))
            except Exception as exp:
                logger.exception(exp)
                raise exp
            logger.error(str("No data buffer for " + self.obj.pathname()))
            return self.obj.pathname()
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                                            aws_secret_access_key=self.obj.store.get_secret(), \
                                            aws_access_key_id=self.obj.store.get_access(), \
                                            endpoint_url=self.obj.store.get_endpoint()) as client:
            try:
                with async_timeout.timeout(10):
                    put_obj = await client.put_object(Bucket=self.obj.bucketname(), \
                                                    Key=self.obj.objname(), \
                                                    Body=self.obj.databuf)

                retcode = put_obj['ResponseMetadata']['HTTPStatusCode']
                logger.debug("s3 put : " + self.obj.pathname() + " " + str(retcode))
                return self.obj.pathname()
            except Exception as exp:
                logger.exception(exp)
                raise exp

    async def get(self, session=None):
        """ GET the obj using the S3 protocol ane return the buffer """
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                                            aws_secret_access_key=self.obj.store.get_secret(), \
                                            aws_access_key_id=self.obj.store.get_access(), \
                                            endpoint_url=self.obj.store.get_endpoint()) as client:
            try:
                logger.debug("s3 get : " + self.obj.pathname())
                with async_timeout.timeout(10):
                    gobj = await client.get_object(Bucket=self.obj.bucketname(),\
                                                   Key=self.obj.objname())
                    body = await gobj['Body'].read()
                    gobj['Body'].close()
                    return body
            except Exception as exp:
                logger.exception(exp)
                raise exp

    async def delete(self, session=None):
        """ delete this object """
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                                        aws_secret_access_key=self.obj.store.get_secret(), \
                                        aws_access_key_id=self.obj.store.get_access(), \
                                        endpoint_url=self.obj.store.get_endpoint()) as client:
            try:
                del_obj = await client.delete_object(Bucket=self.obj.bucketname(), \
                                                     Key=self.obj.objname())
                retcode = del_obj['ResponseMetadata']['HTTPStatusCode']
                logger.debug("s3 delete : " + self.obj.pathname() + " " + str(retcode))
                return retcode in (200, 204)
            except Exception as exp:
                logger.exception(exp)
                return False

    async def exists(self, session):
        """ EXISTS checks if the objects is there or not there """
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                                    aws_secret_access_key=self.obj.store.get_secret(), \
                                    aws_access_key_id=self.obj.store.get_access(), \
                                    endpoint_url=self.obj.store.get_endpoint()) as client:
            try:
                hd_obj = await client.head_object(Bucket=self.obj.bucketname(),\
                                              Key=self.obj.objname())
                retcode = hd_obj['ResponseMetadata']['HTTPStatusCode']
                logger.debug("s3 exists? : " + self.obj.pathname() + " " + str(retcode))
                return retcode == 200
            except:
                return False

    async def info(self, session=None):
        """ Retrieve the Meta data that exists on this object """
        async with session.create_client('s3', region_name=self.obj.store.get_region(), \
                                    aws_secret_access_key=self.obj.store.get_secret(), \
                                    aws_access_key_id=self.obj.store.get_access(), \
                                    endpoint_url=self.obj.store.get_endpoint()) as client:
            try:
                hd_obj = await client.head_object(Bucket=self.obj.bucketname(),\
                                              Key=self.obj.objname())
                retcode = hd_obj['ResponseMetadata']['HTTPStatusCode']
                logger.debug("s3 info : " + self.obj.pathname() + " " + str(retcode))
                if retcode == 200:
                    return hd_obj['ResponseMetadata']['HTTPHeaders']
                return None
            except Exception as exp:
                logger.exception(exp)
                return None



''' This is primarily for errors and exceptions '''

from xml.etree.ElementTree import fromstring as parse_xml, ParseError

class EdgexException(Exception):
    """Base for errors, exceptions returned by edgex_access components """
    @staticmethod
    def from_bytes(status, body):
        """
            extract the error from the xml response
        """
        if not body:
            raise RuntimeError("HTTP Error {}".format(status))
        try:
            xml = parse_xml(body)
        except ParseError:
            raise RuntimeError(body)
        code_el = xml.find("Code")
        if code_el is None or not code_el.text:
            raise RuntimeError(body)
        class_name = code_el.text
        try:
            cls = globals()[class_name]
        except KeyError:
            raise RuntimeError("Error {} is unknown".format(class_name))
        msg = xml.find("Message")
        return cls(class_name if msg is None else msg.text)

class AccessDenied(EdgexException):
    """Access is Denied"""
    pass
class AccountProblem(EdgexException):
    """Account has a problem"""
    pass
class AmbiguousGrantByEmailAddress(EdgexException):
    """ Grant looks Ambigious """
    pass
class BadDigest(EdgexException):
    """ Bad digest """
    pass
class BucketAlreadyExists(EdgexException):
    """ Bucket already exists  """
    pass
class BucketAlreadyOwnedByYou(EdgexException):
    """ Bucket is already owned by the same user """
    pass
class BucketNotEmpty(EdgexException):
    """ Bucket is not empty """
    pass
class CredentialsNotSupported(EdgexException):
    """ Credentials is not supported  """
    pass
class CrossLocationLoggingProhibited(EdgexException):
    """ Cross location logging is not allowed  """
    pass
class EntityTooSmall(EdgexException):
    """ Too small entity """
    pass
class EntityTooLarge(EdgexException):
    """ Too large entity  """
    pass
class ExpiredToken(EdgexException):
    """ Token has expired """
    pass
class IllegalVersioningConfigurationException(EdgexException):
    """ Version configuration is illegal """
    pass
class IncompleteBody(EdgexException):
    """ Incomplete body """
    pass
class IncorrectNumberOfFilesInPostRequest(EdgexException):
    """ Incorret number of files in the post request """
    pass
class InlineDataTooLarge(EdgexException):
    """ Inline data is too large """
    pass
class InternalError(EdgexException):
    """ Internal Error """
    pass
class InvalidAccessKeyId(EdgexException):
    """ Invalid Access  """
    pass
class InvalidAddressingHeader(EdgexException):
    """ Invalid Address header """
    pass
class InvalidArgument(EdgexException):
    """ Invalid Argument """
    pass
class InvalidBucketName(EdgexException):
    """ Invalid bucket name """
    pass
class InvalidBucketState(EdgexException):
    """ Invalid bucket state """
    pass
class InvalidDigest(EdgexException):
    """ Invalid digest """
    pass
class InvalidEncryptionAlgorithmError(EdgexException):
    """ Invalid encryption algorithm error """
    pass
class InvalidLocationConstraint(EdgexException):
    """ Invalid location constraint """
    pass
class InvalidObjectState(EdgexException):
    """ Invalid Object state """
    pass
class InvalidPart(EdgexException):
    """ Invalid part """
    pass
class InvalidPartOrder(EdgexException):
    """ Invalid part order """
    pass
class InvalidPayer(EdgexException):
    """ Invalid Player """
    pass
class InvalidPolicyDocument(EdgexException):
    """ Invalid policy """
    pass
class InvalidRange(EdgexException):
    """ Invalid range """
    pass
class InvalidRequest(EdgexException):
    """ Invalid request """
    pass
class InvalidSecurity(EdgexException):
    """ Invalid security """
    pass
class InvalidSOAPRequest(EdgexException):
    """ Invalid SOAP """
    pass
class InvalidStorageClass(EdgexException):
    """ Invalid Storage class """
    pass
class InvalidTargetBucketForLogging(EdgexException):
    """ Invalid target bucket """
    pass
class InvalidToken(EdgexException):
    """ Token is invalid """
    pass
class InvalidURI(EdgexException):
    """ URI is invalid """
    pass
class InvalidCommand(EdgexException):
    """ Command is invalid """
    pass
class InvalidStore(EdgexException):
    """ Store is invalid """
    pass
class InvalidDataBuffer(EdgexException):
    """ Buffer is invalid """
    pass
class KeyTooLong(EdgexException):
    """ Key is too long """
    pass
class MalformedACLError(EdgexException):
    """ACL is malformed  """
    pass
class MalformedPOSTRequest(EdgexException):
    """ POST request is malformed """
    pass
class MalformedXML(EdgexException):
    """ XML is malformed """
    pass
class MaxMessageLengthExceeded(EdgexException):
    """ Message length has exceeded the max """
    pass
class MaxPostPreDataLengthExceededError(EdgexException):
    """ POST predata length exceeded the error """
    pass
class MetadataTooLarge(EdgexException):
    """ Too large metadata """
    pass
class MethodNotAllowed(EdgexException):
    """ Method is not allowed """
    pass
class MissingAttachment(EdgexException):
    """ Attachment is missing  """
    pass
class MissingContentLength(EdgexException):
    """ COntent length is missing """
    pass
class MissingRequestBodyError(EdgexException):
    """ Request body is missing  """
    pass
class MissingSecurityElement(EdgexException):
    """ Security element is missing """
    pass
class MissingSecurityHeader(EdgexException):
    """ Secirity header is missing """
    pass
class NoLoggingStatusForKey(EdgexException):
    """ No logging status for this key """
    pass
class NoSuchBucket(EdgexException):
    """ No such bucket """
    pass
class NoSuchKey(EdgexException):
    """ No such key """
    pass
class NoSuchLifecycleConfiguration(EdgexException):
    """ No such life cycle configuration  """
    pass
class NoSuchUpload(EdgexException):
    """ No suck upload """
    pass
class NoSuchVersion(EdgexException):
    """ No such version  """
    pass
class NotSignedUp(EdgexException):
    """ Not signed up """
    pass
class NotSuchBucketPolicy(EdgexException):
    """ No such bucket """
    pass
class OperationAborted(EdgexException):
    """ Operation aborted """
    pass
class PermanentRedirect(EdgexException):
    """ Permanent redirect """
    pass
class PreconditionFailed(EdgexException):
    """ Precondition failed """
    pass
class Redirect(EdgexException):
    """ Redirect """
    pass
class RestoreAlreadyInProgress(EdgexException):
    """ Restore is already in progress """
    pass
class RequestIsNotMultiPartContent(EdgexException):
    """ Request is not multi-part content """
    pass
class RequestTimeout(EdgexException):
    """ Request timeout """
    pass
class RequestTimeTooSkewed(EdgexException):
    """ Request time is skewed """
    pass
class RequestTorrentOfBucketError(EdgexException):
    """ Request torrent bucket error """
    pass
class SignatureDoesNotMatch(EdgexException):
    """ Signature mis-match  """
    pass
class ServiceUnavailable(EdgexException):
    """ Service is unavailable """
    pass
class SlowDown(EdgexException):
    """ Slow down """
    pass
class TemporaryRedirect(EdgexException):
    """ Temporary redirect """
    pass
class TokenRefreshRequired(EdgexException):
    """ Token needs to be refreshed """
    pass
class TooManyBuckets(EdgexException):
    """ Too many buckets """
    pass
class UnexpectedContent(EdgexException):
    """ Unexpected content """
    pass
class UnresolvableGrantByEmailAddress(EdgexException):
    """ Unresolved Grant by email  """
    pass
class UserKeyMustBeSpecified(EdgexException):
    """ User key must be specified """
    pass
class EmptyTag(EdgexException):
    """ Tag is empty """
    pass



''' Filesystem specific access '''

import os
import aiofiles

from logzero import logger
from .edgex_base import EdgexStoreBase, EdgexAccessBase
from .edgex_base import EdgexStoreBase, EdgexAccessBase
from .edgex_exceptions import *

MAX_SINGLE_OBJ = 5* 1024 * 1024 * 1024 # 5Gb

class EdgexFSStore(EdgexStoreBase):
    """ Stuff needed to access the local filesystem
    """
    def __init__(self, cfg):
        super().__init__(cfg)
        self.cwd = os.getcwd()
        self.endpoint = cfg['ENDPOINT']

    def basename(self):
        """ Get the basename """
        return os.path.basename(self.cwd)

    def get_endpoint(self):
        """ Get the endpoint """
        return self.endpoint

    def change_value(self, key, value):
        """ change config value """
        if key.upper() == "TAG":
            super().set_tag(value)

class EdgexFSAccess(EdgexAccessBase):
    ''' functions for filesystem access '''
    def __init__(self, obj):
        super().__init__(obj, "FS")

    async def list(self, session=None):
        ''' object listing or directory listing '''
        logger.debug("fs list " + self.obj.pathname())
        final_list = []
        if self.obj.isfolder():
            final_list = os.listdir(self.obj.pathname())
            i = 0
            for f_l in final_list:
                if os.path.isdir(self.obj.pathname() + "/" + f_l):
                    final_list[i] = f_l + "/"
                i += 1
        else:
            if os.path.isfile(self.obj.pathname()):
                final_list.append(self.obj.pathname())
        return final_list

    async def put(self, session=None):
        ''' write out an object '''
        logger.debug("fs put " + self.obj.pathname())
        isdbuf = (self.obj.databuf is not None)

        logger.info(str("put " + self.obj.pathname() + " databuf " + str(isdbuf)))

        if not isdbuf:
            try:
                os.makedirs(os.path.dirname(self.obj.pathname()))
            except Exception as exp:
                logger.exception(exp)
                raise exp
            logger.error(str("No data buffer for " + self.obj.pathname()))
            return self.obj.pathname()
        if not os.path.exists(os.path.dirname(self.obj.pathname())):
            try:
                os.makedirs(os.path.dirname(self.obj.pathname()))
            except Exception as exp:
                logger.exception(exp)
                raise exp
        async with aiofiles.open(self.obj.pathname(), mode='wb') as f_l:
            await f_l.write(self.obj.databuf)
            await f_l.flush()
        return self.obj.pathname()

    async def get(self, session=None):
        ''' read an object '''
        logger.debug("fs get " + self.obj.pathname())
        file_size = os.stat(self.obj.pathname()).st_size
        if file_size > MAX_SINGLE_OBJ:
            raise EntityTooLarge(str(file_size))
        async with aiofiles.open(self.obj.pathname(), mode='rb') as f_l:
            file_data = await f_l.read()
        f_l.close()
        return file_data

    async def delete(self, session=None):
        ''' delete an object '''
        logger.debug("fs delete " + self.obj.pathname())
        if os.path.isfile(self.obj.pathname()):
            os.remove(self.obj.pathname())
            return True
        if os.path.isdir(self.obj.pathname()):
            dentries = os.listdir(self.obj.pathname())
            if len(dentries) == 0:
                os.rmdir(self.obj.pathname())

    async def exists(self, session=None):
        ''' check of the file exists '''
        logger.debug("fs exists " + self.obj.pathname())
        return self.obj.stat()

    async def info(self, session=None):
        ''' meta info on the object if any '''
        logger.debug("fs info " + self.obj.pathname())
        if self.obj.stat() is True:
            metadata = {self.obj.pathname():os.stat(self.obj.pathname())}
            return metadata
        return None



''' place objects in out of only memory '''

from .edgex_exceptions import *
from .edgex_base import EdgexStoreBase, EdgexAccessBase



class EdgexMemStore(EdgexStoreBase):
    """ store stuff in memory for small things """
    def __init__(self, cfg):
        super().__init__(cfg)
        self.bucketname = cfg['BUCKET']
        self.endpoint = cfg['ENDPOINT']
    def get_bucketname(self):
        """ the mem space is denoted by a bucket name """
        return self.bucketname
    def get_endpoint(self):
        """ get the endpoint """
        return self.endpoint

class EdgexMemAccess(EdgexAccessBase):
    """ In-memory access """
    def __init__(self, obj):
        super().__init__(obj, "MEM")

    async def list(self, session=None):
        """ List the element in-memory """
        pass
    async def put(self, session=None):
        """ put the databuf of the object in-memory """
        pass
    async def get(self, session=None):
        """ get the databuf from in-memory """
        if not self.obj.databuf:
            raise InvalidDataBuffer(self.obj.pathname())
        return self.obj.databuf
    async def delete(self, session=None):
        """ delete the entry from in-memory """
        if not self.obj.databuf:
            raise InvalidDataBuffer(self.obj.pathname())
        self.obj.databuf = ""
    async def exists(self, session=None):
        """ check of this exists in-memory """
        pass
    async def info(self, session=None):
        """ return the meta info on this """
        pass



"""
See:
    https://packaging.python.org/en/latest/distributing.html
    https://github.com/pypa/sampleproject
    https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/

"""

import sys
import os

import os.path as path

from setuptools import setup

if sys.version_info < (3, 0, 0):
    sys.stdout.write("Need Python3 or above. Detected: %s\n" \
                     % sys.version.split(None, 1)[0])
    raise RuntimeError("edgex_access is for Python 3")

HERE = path.abspath(path.dirname(__file__))
VPATH = os.path.join(HERE, 'edgex_access', 'version.py')
__version__ = eval(open(VPATH).read())


###################################################################
LONG_DESCRIPTION = "edgex_access: a Python library for accessing AWS using the S3 protocol"

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='edgex_access',
    version=__version__,
    description="S3 protocol Data access to AWS S3",
    long_description=LONG_DESCRIPTION,
    url="http://www.github.com/nacharya/edgex_access",
    author="Nabin Acharya",
    author_email="nabin.acharya@gmail.com",
    license="MIT",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    scripts=['sp3/sp3'],
    keywords='requests edgex_access s3 scaleout store distributed',
    packages=['edgex_access'],
    python_requires='>=3',
    install_requires=['urllib3', 'requests_aws4auth', 'aiobotocore', \
                      'simplejson', 'lxml', 'asyncio'],
    include_package_data=True,
    project_urls={
        'Bug Reports': 'https://github.com/nacharya/edgex_access/issues',
        'Source': 'https://github.com/nacharya/edgex_access/',
    },
    zip_safe=False
)



''' Google Storage specific using google APIs '''

import os
from .edgex_base import EdgexStoreBase, EdgexAccessBase

class EdgexGoogStore(EdgexStoreBase):
    """ Stuff needed to access the local filesystem
    """
    def __init__(self, cfg):
        super().__init__(cfg)
        self.cwd = os.getcwd()
    def basename(self):
        """ Get the basename """
        return os.path.basename(self.cwd)
    def get_endpoint(self):
        """ Get the endpoint """
        return self.basename()


class EdgexGoogAccess(EdgexAccessBase):
    ''' Google specific storage API access '''
    pass
